"""Declarations for Python renderings of basic JavaScript types."""

from __future__ import annotations

from abc import ABC
from collections.abc import Generator, MutableMapping, MutableSequence
from datetime import datetime
from typing import Any, TypeAlias


class JSUndefinedType:
    """The JavaScript undefined type.

    Where JavaScript null is represented as None, undefined is represented as this
    type."""

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "JSUndefined"


JSUndefined = JSUndefinedType()


class JSObject(ABC):  # noqa: B024
    """A JavaScript object."""


class JSMappedObject(
    MutableMapping["PythonJSConvertedTypes", "PythonJSConvertedTypes"], JSObject
):
    """A JavaScript object with Pythonic MutableMapping methods (`keys()`,
    `__getitem__()`, etc).

    `keys()` and `__iter__()` will return properties from any prototypes as well as this
    object, as if using a for-in statement in JavaScript.
    """


class JSArray(MutableSequence["PythonJSConvertedTypes"], JSObject):
    """JavaScript array.

    Has Pythonic MutableSequence methods (e.g., `insert()`, `__getitem__()`, ...).
    """


class JSFunction(JSMappedObject):
    """JavaScript function.

    You can call this object from Python, passing in positional args to match what the
    JavaScript function expects, along with a keyword argument, `timeout_sec`.
    """

    def __call__(
        self,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        raise NotImplementedError


class JSSymbol(JSMappedObject):
    """JavaScript symbol."""


class JSPromise(JSObject):
    """JavaScript Promise.

    To get a value, call `promise.get()` to block, or `await promise` from within an
    `async` coroutine. Either will raise a Python exception if the JavaScript Promise
    is rejected.
    """

    def get(self, *, timeout: float | None = None) -> PythonJSConvertedTypes:
        raise NotImplementedError

    def __await__(self) -> Generator[Any, None, Any]:
        raise NotImplementedError

    async def _do_await(self) -> PythonJSConvertedTypes:
        raise NotImplementedError


PythonJSConvertedTypes: TypeAlias = (
    JSUndefinedType
    | bool
    | int
    | float
    | str
    | JSObject
    | datetime
    | memoryview
    | JSPromise
    | JSFunction
    | JSMappedObject
    | JSSymbol
    | JSArray
    | None
)



from __future__ import annotations

from asyncio import (
    FIRST_COMPLETED,
    Task,
    create_task,
    get_running_loop,
    wait,
)
from collections.abc import AsyncIterator, Awaitable, Callable, Iterator
from contextlib import asynccontextmanager, contextmanager, suppress
from itertools import count
from traceback import format_exc
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
)

from py_mini_racer._abstract_context import AbstractContext
from py_mini_racer._dll import init_mini_racer, mr_callback_func
from py_mini_racer._exc import JSEvalException
from py_mini_racer._sync_future import SyncFuture
from py_mini_racer._types import (
    JSArray,
    JSFunction,
    JSObject,
    JSPromise,
    JSUndefined,
    JSUndefinedType,
    PythonJSConvertedTypes,
)
from py_mini_racer._value_handle import (
    ValueHandle,
    python_to_value_handle,
)

if TYPE_CHECKING:
    import ctypes
    from asyncio import Future

    from py_mini_racer._abstract_context import AbstractValueHandle
    from py_mini_racer._value_handle import RawValueHandleType

PyJsFunctionType = Callable[..., Awaitable[PythonJSConvertedTypes]]
AsyncCleanupType = Callable[[], Awaitable[None]]


def context_count() -> int:
    """For tests only: how many context handles are still allocated?"""

    dll = init_mini_racer(ignore_duplicate_init=True)
    return int(dll.mr_context_count())


class _CallbackRegistry:
    def __init__(
        self,
        raw_handle_wrapper: Callable[[RawValueHandleType], AbstractValueHandle],
    ) -> None:
        self._active_callbacks: dict[
            int,
            Callable[[PythonJSConvertedTypes | JSEvalException], None],
        ] = {}

        # define an all-purpose callback:
        @mr_callback_func
        def mr_callback(callback_id: int, raw_val_handle: RawValueHandleType) -> None:
            val_handle = raw_handle_wrapper(raw_val_handle)
            callback = self._active_callbacks[callback_id]
            callback(val_handle.to_python())

        self.mr_callback = mr_callback

        self._next_callback_id = count()

    def register(
        self,
        func: Callable[[PythonJSConvertedTypes | JSEvalException], None],
    ) -> int:
        callback_id = next(self._next_callback_id)

        self._active_callbacks[callback_id] = func

        return callback_id

    def cleanup(self, callback_id: int) -> None:
        self._active_callbacks.pop(callback_id)


class Context(AbstractContext):
    """Wrapper for all operations involving the DLL and C++ MiniRacer::Context."""

    def __init__(
        self,
        dll: ctypes.CDLL,
    ) -> None:
        self._dll: ctypes.CDLL | None = dll

        self._callback_registry = _CallbackRegistry(self._wrap_raw_handle)
        self._ctx = dll.mr_init_context(self._callback_registry.mr_callback)

    def _get_dll(self) -> ctypes.CDLL:
        if self._dll is None:
            msg = "Operation on closed Context"
            raise ValueError(msg)

        return self._dll

    def v8_version(self) -> str:
        return str(self._get_dll().mr_v8_version().decode("utf-8"))

    def v8_is_using_sandbox(self) -> bool:
        """Checks for enablement of the V8 Sandbox. See https://v8.dev/blog/sandbox."""

        return bool(self._get_dll().mr_v8_is_using_sandbox())

    def evaluate(
        self,
        code: str,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        code_handle = python_to_value_handle(self, code)

        with self._run_mr_task(
            self._get_dll().mr_eval,
            self._ctx,
            code_handle.raw,
        ) as future:
            return future.get(timeout=timeout_sec)

    def promise_then(
        self,
        promise: JSPromise,
        on_resolved: JSFunction,
        on_rejected: JSFunction,
    ) -> None:
        promise_handle = python_to_value_handle(self, promise)
        then_name_handle = python_to_value_handle(self, "then")
        then_func = self._wrap_raw_handle(
            self._get_dll().mr_get_object_item(
                self._ctx,
                promise_handle.raw,
                then_name_handle.raw,
            ),
        ).to_python_or_raise()

        then_func = cast("JSFunction", then_func)
        then_func(on_resolved, on_rejected, this=promise)

    def get_identity_hash(self, obj: JSObject) -> int:
        obj_handle = python_to_value_handle(self, obj)

        ret = self._wrap_raw_handle(
            self._get_dll().mr_get_identity_hash(self._ctx, obj_handle.raw),
        ).to_python_or_raise()
        return cast("int", ret)

    def get_own_property_names(
        self,
        obj: JSObject,
    ) -> tuple[PythonJSConvertedTypes, ...]:
        obj_handle = python_to_value_handle(self, obj)

        names = self._wrap_raw_handle(
            self._get_dll().mr_get_own_property_names(self._ctx, obj_handle.raw),
        ).to_python_or_raise()
        if not isinstance(names, JSArray):
            raise TypeError
        return tuple(names)

    def get_object_item(
        self,
        obj: JSObject,
        key: PythonJSConvertedTypes,
    ) -> PythonJSConvertedTypes:
        obj_handle = python_to_value_handle(self, obj)
        key_handle = python_to_value_handle(self, key)

        return self._wrap_raw_handle(
            self._get_dll().mr_get_object_item(
                self._ctx,
                obj_handle.raw,
                key_handle.raw,
            ),
        ).to_python_or_raise()

    def set_object_item(
        self,
        obj: JSObject,
        key: PythonJSConvertedTypes,
        val: PythonJSConvertedTypes,
    ) -> None:
        obj_handle = python_to_value_handle(self, obj)
        key_handle = python_to_value_handle(self, key)
        val_handle = python_to_value_handle(self, val)

        # Convert the value just to convert any exceptions (and GC the result)
        self._wrap_raw_handle(
            self._get_dll().mr_set_object_item(
                self._ctx,
                obj_handle.raw,
                key_handle.raw,
                val_handle.raw,
            ),
        ).to_python_or_raise()

    def del_object_item(self, obj: JSObject, key: PythonJSConvertedTypes) -> None:
        obj_handle = python_to_value_handle(self, obj)
        key_handle = python_to_value_handle(self, key)

        # Convert the value just to convert any exceptions (and GC the result)
        self._wrap_raw_handle(
            self._get_dll().mr_del_object_item(
                self._ctx,
                obj_handle.raw,
                key_handle.raw,
            ),
        ).to_python_or_raise()

    def del_from_array(self, arr: JSArray, index: int) -> None:
        arr_handle = python_to_value_handle(self, arr)

        # Convert the value just to convert any exceptions (and GC the result)
        self._wrap_raw_handle(
            self._get_dll().mr_splice_array(self._ctx, arr_handle.raw, index, 1, None),
        ).to_python_or_raise()

    def array_insert(
        self,
        arr: JSArray,
        index: int,
        new_val: PythonJSConvertedTypes,
    ) -> None:
        arr_handle = python_to_value_handle(self, arr)
        new_val_handle = python_to_value_handle(self, new_val)

        # Convert the value just to convert any exceptions (and GC the result)
        self._wrap_raw_handle(
            self._get_dll().mr_splice_array(
                self._ctx,
                arr_handle.raw,
                index,
                0,
                new_val_handle.raw,
            ),
        ).to_python_or_raise()

    def call_function(
        self,
        func: JSFunction,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        argv = cast("JSArray", self.evaluate("[]"))
        for arg in args:
            argv.append(arg)

        func_handle = python_to_value_handle(self, func)
        this_handle = python_to_value_handle(self, this)
        argv_handle = python_to_value_handle(self, argv)

        with self._run_mr_task(
            self._get_dll().mr_call_function,
            self._ctx,
            func_handle.raw,
            this_handle.raw,
            argv_handle.raw,
        ) as future:
            return future.get(timeout=timeout_sec)

    def set_hard_memory_limit(self, limit: int) -> None:
        self._get_dll().mr_set_hard_memory_limit(self._ctx, limit)

    def set_soft_memory_limit(self, limit: int) -> None:
        self._get_dll().mr_set_soft_memory_limit(self._ctx, limit)

    def was_hard_memory_limit_reached(self) -> bool:
        return bool(self._get_dll().mr_hard_memory_limit_reached(self._ctx))

    def was_soft_memory_limit_reached(self) -> bool:
        return bool(self._get_dll().mr_soft_memory_limit_reached(self._ctx))

    def low_memory_notification(self) -> None:
        self._get_dll().mr_low_memory_notification(self._ctx)

    def heap_stats(self) -> str:
        with self._run_mr_task(self._get_dll().mr_heap_stats, self._ctx) as future:
            return cast("str", future.get())

    def heap_snapshot(self) -> str:
        """Return a snapshot of the V8 isolate heap."""

        with self._run_mr_task(self._get_dll().mr_heap_snapshot, self._ctx) as future:
            return cast("str", future.get())

    def value_count(self) -> int:
        """For tests only: how many value handles are still allocated?"""

        return int(self._get_dll().mr_value_count(self._ctx))

    @contextmanager
    def js_callback(
        self,
        func: Callable[[PythonJSConvertedTypes | JSEvalException], None],
    ) -> Iterator[JSFunction]:
        """Make a JS callback which forwards to the given Python function.

        Note that it's crucial that the given Python function *not* call back
        into the C++ MiniRacer context, or it will deadlock. Instead it should
        signal another thread; e.g., by putting received data onto a queue or
        future.
        """

        callback_id = self._callback_registry.register(func)

        cb = self._wrap_raw_handle(
            self._get_dll().mr_make_js_callback(self._ctx, callback_id),
        )
        cb_py = cast("JSFunction", cb.to_python_or_raise())

        yield cb_py

        self._callback_registry.cleanup(callback_id)

    def _wrap_raw_handle(self, raw: RawValueHandleType) -> ValueHandle:
        return ValueHandle(self, raw)

    def create_intish_val(self, val: int, typ: int) -> AbstractValueHandle:
        return self._wrap_raw_handle(
            self._get_dll().mr_alloc_int_val(
                self._ctx,
                val,
                typ,
            ),
        )

    def create_doublish_val(self, val: float, typ: int) -> AbstractValueHandle:
        return self._wrap_raw_handle(
            self._get_dll().mr_alloc_double_val(
                self._ctx,
                val,
                typ,
            ),
        )

    def create_string_val(self, val: str, typ: int) -> AbstractValueHandle:
        b = val.encode("utf-8")
        return self._wrap_raw_handle(
            self._get_dll().mr_alloc_string_val(
                self._ctx,
                b,
                len(b),
                typ,
            ),
        )

    def free(self, val_handle: AbstractValueHandle) -> None:
        dll = self._dll
        if dll is not None:
            dll.mr_free_value(self._ctx, val_handle.raw)

    @contextmanager
    def _run_mr_task(self, dll_method: Any, *args: Any) -> Iterator[SyncFuture]:  # noqa: ANN401
        """Manages those tasks which generate callbacks from the MiniRacer DLL.

        Several MiniRacer functions (JS evaluation and 2 heap stats calls) are
        asynchronous. They take a function callback and callback data parameter, and
        they return a task handle.

        In this method, we create a future for each callback to get the right data to
        the right caller, and we manage the lifecycle of the task and task handle.
        """

        future = SyncFuture()

        def callback(value: PythonJSConvertedTypes | JSEvalException) -> None:
            if isinstance(value, JSEvalException):
                future.set_exception(value)
            else:
                future.set_result(value)

        callback_id = self._callback_registry.register(callback)

        # Start the task:
        task_id = dll_method(*args, callback_id)
        try:
            # Let the caller handle waiting on the result:
            yield future
        finally:
            # Cancel the task if it's not already done (this call is ignored if it's
            # already done)
            self._get_dll().mr_cancel_task(self._ctx, task_id)

            # If the caller gives up on waiting, let's at least await the
            # cancelation error for GC purposes:
            with suppress(Exception):
                future.get()

            self._callback_registry.cleanup(callback_id)

    @asynccontextmanager
    async def wrap_py_function(
        self,
        func: PyJsFunctionType,
    ) -> AsyncIterator[JSFunction]:
        async def run_one(params: JSArray) -> None:
            arguments, resolve, reject = params
            arguments = cast("JSArray", arguments)
            resolve = cast("JSFunction", resolve)
            reject = cast("JSFunction", reject)
            try:
                result = await func(*arguments)
                resolve(result)
            except Exception:  # noqa: BLE001
                # Convert this Python exception into a JS exception so we can send it
                # into JS:
                s = f"Error running Python function:\n{format_exc()}"
                err_maker = self.evaluate("s => new Error(s)")
                err_maker = cast("JSFunction", err_maker)
                err = err_maker(s)
                reject(err)

        pending: set[Task[PythonJSConvertedTypes | JSEvalException] | Future[bool]] = (
            set()
        )

        def process(params: PythonJSConvertedTypes | JSEvalException) -> None:
            params = cast("JSArray", params)

            # Start a new task to run the new task:
            task = create_task(run_one(params))
            pending.add(task)

        loop = get_running_loop()

        def on_called(value: PythonJSConvertedTypes | JSEvalException) -> None:
            loop.call_soon_threadsafe(process, value)

        async def await_pending() -> None:
            nonlocal pending
            while pending:
                done, pending = await wait(pending, return_when=FIRST_COMPLETED)
                for coro in done:
                    await coro

        shutdown: Future[bool] = loop.create_future()
        pending.add(shutdown)
        pending_awaiter = create_task(await_pending())
        try:
            with self.js_callback(on_called) as callback:
                wrapper = self.evaluate(
                    """
callback => {
    return (...arguments) => {
        let p = Promise.withResolvers();

        callback(arguments, p.resolve, p.reject);

        return p.promise;
    }
}
""",
                )
                wrapper = cast("JSFunction", wrapper)

                wrapped = wrapper(callback)
                wrapped = cast("JSFunction", wrapped)

                yield wrapped
        finally:
            # Stop accepting calls:
            shutdown.set_result(True)
            await pending_awaiter

    def close(self) -> None:
        dll, self._dll = self._dll, None
        if dll:
            dll.mr_free_context(self._ctx)

    def __del__(self) -> None:
        self.close()



from __future__ import annotations

import json
from json import JSONEncoder
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
)

from py_mini_racer._context import Context
from py_mini_racer._dll import init_mini_racer
from py_mini_racer._exc import MiniRacerBaseException
from py_mini_racer._set_timeout import INSTALL_SET_TIMEOUT

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager
    from types import TracebackType

    from typing_extensions import Self

    from py_mini_racer._context import PyJsFunctionType
    from py_mini_racer._types import JSFunction, PythonJSConvertedTypes


class WrongReturnTypeException(MiniRacerBaseException):
    """Invalid type returned by the JavaScript runtime."""

    def __init__(self, typ: type) -> None:
        super().__init__(f"Unexpected return value type {typ}")


class MiniRacer:
    """
    MiniRacer evaluates JavaScript code using a V8 isolate.

    A MiniRacer instance can be explicitly closed using the close() method, or by using
    the MiniRacer as a context manager, i.e,:

    with MiniRacer() as mr:
        ...

    The MiniRacer instance will otherwise clean up the underlying V8 resource upon
    garbage collection.

    Attributes:
        json_impl: JSON module used by helper methods default is
            [json](https://docs.python.org/3/library/json.html)
    """

    json_impl: ClassVar[Any] = json

    def __init__(self) -> None:
        dll = init_mini_racer(ignore_duplicate_init=True)

        self._ctx = Context(dll)

        self.eval(INSTALL_SET_TIMEOUT)

    def close(self) -> None:
        """Close this MiniRacer instance.

        It is an error to use this MiniRacer instance or any JS objects returned by it
        after calling this method.
        """
        self._ctx.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        del exc_type
        del exc_val
        del exc_tb
        self.close()

    @property
    def v8_version(self) -> str:
        """Return the V8 version string."""
        return self._ctx.v8_version()

    def eval(
        self,
        code: str,
        timeout: float | None = None,
        timeout_sec: float | None = None,
        max_memory: int | None = None,
    ) -> PythonJSConvertedTypes:
        """Evaluate JavaScript code in the V8 isolate.

        Side effects from the JavaScript evaluation is persisted inside a context
        (meaning variables set are kept for the next evaluation).

        The JavaScript value returned by the last expression in `code` is converted to
        a Python value and returned by this method. Only primitive types are supported
        (numbers, strings, buffers...). Use the
        [py_mini_racer.MiniRacer.execute][] method to return more complex
        types such as arrays or objects.

        The evaluation can be interrupted by an exception for several reasons: a limit
        was reached, the code could not be parsed, a returned value could not be
        converted to a Python value.

        Args:
            code: JavaScript code
            timeout: number of milliseconds after which the execution is interrupted.
                This is deprecated; use timeout_sec instead.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted.
        """

        if max_memory is not None:
            self.set_hard_memory_limit(max_memory)

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        return self._ctx.evaluate(code=code, timeout_sec=timeout_sec)

    def execute(
        self,
        expr: str,
        timeout: float | None = None,
        timeout_sec: float | None = None,
        max_memory: int | None = None,
    ) -> Any:  # noqa: ANN401
        """Helper to evaluate a JavaScript expression and return composite types.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using `json_impl`.

        Args:
            expr: JavaScript expression
            timeout: number of milliseconds after which the execution is interrupted.
                This is deprecated; use timeout_sec instead.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted.
        """

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        wrapped_expr = f"JSON.stringify((function(){{return ({expr})}})())"
        ret = self.eval(wrapped_expr, timeout_sec=timeout_sec, max_memory=max_memory)
        if not isinstance(ret, str):
            raise WrongReturnTypeException(type(ret))
        return self.json_impl.loads(ret)

    def call(
        self,
        expr: str,
        *args: Any,  # noqa: ANN401
        encoder: type[JSONEncoder] | None = None,
        timeout: float | None = None,
        timeout_sec: float | None = None,
        max_memory: int | None = None,
    ) -> Any:  # noqa: ANN401
        """Helper to call a JavaScript function and return compositve types.

        The `expr` argument refers to a JavaScript function in the current V8
        isolate context. Further positional arguments are serialized using the JSON
        implementation `json_impl` and passed to the JavaScript function as arguments.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using `json_impl`.

        Args:
            expr: JavaScript expression referring to a function
            encoder: Custom JSON encoder
            timeout: number of milliseconds after which the execution is
                interrupted.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted
        """

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        json_args = self.json_impl.dumps(args, separators=(",", ":"), cls=encoder)
        js = f"{expr}.apply(this, {json_args})"
        return self.execute(js, timeout_sec=timeout_sec, max_memory=max_memory)

    def wrap_py_function(
        self,
        func: PyJsFunctionType,
    ) -> AbstractAsyncContextManager[JSFunction]:
        """Wrap a Python function such that it can be called from JS.

        To be wrapped and exposed in JavaScript, a Python function should:

          1. Be async,
          2. Accept variable positional arguments each of type PythonJSConvertedTypes,
             and
          3. Return one value of type PythonJSConvertedTypes (a type union which
             includes None).

        The function is rendered on the JavaScript side as an async function (i.e., a
        function which returns a Promise).

        Returns:
            An async context manager which, when entered, yields a JS Function which
            can be passed into MiniRacer and called by JS code.
        """

        return self._ctx.wrap_py_function(func)

    def set_hard_memory_limit(self, limit: int) -> None:
        """Set a hard memory limit on this V8 isolate.

        JavaScript execution will be terminated when this limit is reached.

        :param int limit: memory limit in bytes or 0 to reset the limit
        """
        self._ctx.set_hard_memory_limit(limit)

    def set_soft_memory_limit(self, limit: int) -> None:
        """Set a soft memory limit on this V8 isolate.

        The Garbage Collection will use a more aggressive strategy when
        the soft limit is reached but the execution will not be stopped.

        :param int limit: memory limit in bytes or 0 to reset the limit
        """
        self._ctx.set_soft_memory_limit(limit)

    def was_hard_memory_limit_reached(self) -> bool:
        """Return true if the hard memory limit was reached on the V8 isolate."""
        return self._ctx.was_hard_memory_limit_reached()

    def was_soft_memory_limit_reached(self) -> bool:
        """Return true if the soft memory limit was reached on the V8 isolate."""
        return self._ctx.was_soft_memory_limit_reached()

    def low_memory_notification(self) -> None:
        """Ask the V8 isolate to collect memory more aggressively."""
        self._ctx.low_memory_notification()

    def heap_stats(self) -> Any:  # noqa: ANN401
        """Return the V8 isolate heap statistics."""

        return self.json_impl.loads(self._ctx.heap_stats())


# Compatibility with versions 0.4 & 0.5
StrictMiniRacer = MiniRacer



"""An implementation of setTimeout and clearTimeout using only the ECMA standard lib.

V8 does not include an implementation of the setTimeout or clearTimeout APIs, because
they are web APIs (defined in the DOM and Web Worker spec), not ECMAScript APIs.
Nonetheless they are hugely useful and generally taken for granted, so let's go ahead
and install a basic implementation.

We base this on the Atomics.waitAsync API, which *is* part of the ECMAScript spec.
(It would be possible for MiniRacer to implement this on the C++ side instead, but
this way seems simpler.)
"""

from __future__ import annotations

INSTALL_SET_TIMEOUT = """
class __TimerManager {
  constructor() {
    this.next_idx = 0;
    this.pending = new Set();
  }

  setTimeout(func, delay) {
    const id = this.next_idx++;

    const shared = new SharedArrayBuffer(8);
    const view = new Int32Array(shared);

    const args = Array.prototype.slice.call(arguments, 2);

    let callback = () => {
        if (this.pending.has(id)) {
            this.pending.delete(id);
            func(...args);
        }
    };

    this.pending.add(id);
    Atomics.waitAsync(view, 0, 0, delay).value.then(() => callback());

    return id;
  }

  clearTimeout(timeout_id) {
    this.pending.delete(timeout_id);
  }
}

var __timer_manager = new __TimerManager();
var setTimeout = (...arguments) => __timer_manager.setTimeout(...arguments);
var clearTimeout = (...arguments) => __timer_manager.clearTimeout(...arguments);
"""



from __future__ import annotations

from gc import collect
from re import match
from time import sleep, time

import pytest

from py_mini_racer import LibAlreadyInitializedError, MiniRacer, init_mini_racer
from py_mini_racer._context import context_count


def test_init() -> None:
    init_mini_racer(ignore_duplicate_init=True)

    with pytest.raises(LibAlreadyInitializedError):
        init_mini_racer()

    init_mini_racer(ignore_duplicate_init=True)


# Unfortunately while init_mini_racer allows changing V8 flags, it's hard to test
# automatically because only the first use of V8 can set flags. We'd need to
# restart Python between tests.
# Here's a manual test:
@pytest.mark.skip("Has to be run manually")
def test_init_flags() -> None:
    from py_mini_racer import (  # noqa: PLC0415
        DEFAULT_V8_FLAGS,
        MiniRacer,
        init_mini_racer,
    )

    init_mini_racer(flags=(*DEFAULT_V8_FLAGS, "--no-use-strict"))
    mr = MiniRacer()
    # this would normally fail in strict JS mode because foo is not declared:
    mr.eval("foo = 4")


def test_version() -> None:
    mr = MiniRacer()
    assert match(r"^\d+\.\d+\.\d+\.\d+$", mr.v8_version) is not None


def test_sandbox() -> None:
    mr = MiniRacer()
    assert mr._ctx.v8_is_using_sandbox()  # noqa: SLF001


def test_del() -> None:
    # Collect any leftover contexts:
    start = time()
    while time() - start < 10 and context_count() != 0:  # noqa: PLR2004
        collect()
        sleep(0.1)

    assert context_count() == 0

    mr = MiniRacer()
    del mr

    start = time()
    while time() - start < 10 and context_count() != 0:  # noqa: PLR2004
        collect()
        sleep(0.1)

    assert context_count() == 0


def test_interrupts_background_task_on_shutdown() -> None:
    with MiniRacer() as mr:
        # Schedule a never-ending background task:
        mr.eval("setTimeout(() => { while (1); }, 1);")
        # Make sure the task starts:
        sleep(0.1)
        # Now at the end of the with statement, we explicitly close the instance, which
        # should terminate the background task (as oppposed to hanging indefinitely
        # here).



"""Tests JSFunctions"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest

from py_mini_racer import (
    JSEvalException,
    JSTimeoutException,
    MiniRacer,
)
from tests.gc_check import assert_no_v8_objects

if TYPE_CHECKING:
    from py_mini_racer import JSArray, JSFunction, JSMappedObject


def test_function() -> None:
    mr = MiniRacer()
    func = cast("JSFunction", mr.eval("(a) => a"))
    assert func(42) == 42  # noqa: PLR2004
    arr = mr.eval("[41, 42]")
    assert list(cast("JSArray", func(arr))) == [41, 42]
    thing = cast(
        "JSMappedObject",
        mr.eval(
            """\
class Thing {
    constructor(a) {
        this.blob = a;
    }

    stuff(extra) {
        return this.blob + extra;
    }
}
new Thing('start');
""",
        ),
    )
    stuff = cast("JSFunction", thing["stuff"])
    assert stuff("end", this=thing) == "startend"

    del func, arr, thing, stuff
    assert_no_v8_objects(mr)


def test_exceptions() -> None:
    mr = MiniRacer()
    func = cast(
        "JSFunction",
        mr.eval(
            """\
function func(a, b, c) {
    throw new Error('asdf');
}
func
""",
        ),
    )

    with pytest.raises(JSEvalException) as exc_info:
        func()

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:2: Error: asdf
    throw new Error('asdf');
    ^

Error: asdf
    at func (<anonymous>:2:11)
"""
    )

    del func, exc_info
    assert_no_v8_objects(mr)


def test_timeout() -> None:
    mr = MiniRacer()

    func = cast("JSFunction", mr.eval("() => { while(1) { } }"))
    with pytest.raises(JSTimeoutException) as exc_info:
        func(timeout_sec=1)

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"

    del func, exc_info
    assert_no_v8_objects(mr)



from __future__ import annotations

from py_mini_racer import MiniRacer
from tests.gc_check import assert_no_v8_objects


def test_heap_stats() -> None:
    mr = MiniRacer()

    assert mr.heap_stats()["used_heap_size"] > 0
    assert mr.heap_stats()["total_heap_size"] > 0

    assert_no_v8_objects(mr)



"""Python wrappers for JavaScript object types."""

from __future__ import annotations

from asyncio import get_running_loop
from contextlib import ExitStack
from operator import index as op_index
from typing import TYPE_CHECKING, Any, cast

from py_mini_racer._exc import JSArrayIndexError, JSPromiseError
from py_mini_racer._sync_future import SyncFuture
from py_mini_racer._types import (
    JSArray,
    JSFunction,
    JSMappedObject,
    JSObject,
    JSPromise,
    JSSymbol,
    JSUndefined,
    JSUndefinedType,
    PythonJSConvertedTypes,
)

if TYPE_CHECKING:
    from asyncio import Future
    from collections.abc import Callable, Generator, Iterator

    from py_mini_racer._abstract_context import AbstractContext, AbstractValueHandle
    from py_mini_racer._exc import JSEvalException


def _get_exception_msg(reason: PythonJSConvertedTypes) -> str:
    if not isinstance(reason, JSMappedObject):
        return str(reason)

    if "stack" in reason:
        return cast("str", reason["stack"])

    return str(reason)


class JSObjectImpl(JSObject):
    """A JavaScript object."""

    def __init__(
        self,
        ctx: AbstractContext,
        handle: AbstractValueHandle,
    ) -> None:
        self._ctx = ctx
        self._handle = handle

    def __hash__(self) -> int:
        return self._ctx.get_identity_hash(self)

    @property
    def raw_handle(self) -> AbstractValueHandle:
        return self._handle


class JSMappedObjectImpl(JSObjectImpl, JSMappedObject):
    """A JavaScript object with Pythonic MutableMapping methods (`keys()`,
    `__getitem__()`, etc).

    `keys()` and `__iter__()` will return properties from any prototypes as well as this
    object, as if using a for-in statement in JavaScript.
    """

    def __iter__(self) -> Iterator[PythonJSConvertedTypes]:
        return iter(self._get_own_property_names())

    def __getitem__(self, key: PythonJSConvertedTypes) -> PythonJSConvertedTypes:
        return self._ctx.get_object_item(self, key)

    def __setitem__(
        self,
        key: PythonJSConvertedTypes,
        val: PythonJSConvertedTypes,
    ) -> None:
        self._ctx.set_object_item(self, key, val)

    def __delitem__(self, key: PythonJSConvertedTypes) -> None:
        self._ctx.del_object_item(self, key)

    def __len__(self) -> int:
        return len(self._get_own_property_names())

    def _get_own_property_names(self) -> tuple[PythonJSConvertedTypes, ...]:
        return self._ctx.get_own_property_names(self)


class JSArrayImpl(JSArray, JSObjectImpl):
    """JavaScript array.

    Has Pythonic MutableSequence methods (e.g., `insert()`, `__getitem__()`, ...).
    """

    def __len__(self) -> int:
        ret = self._ctx.get_object_item(self, "length")
        return cast("int", ret)

    def __getitem__(self, index: int | slice) -> Any:  # noqa: ANN401
        if not isinstance(index, int):
            raise TypeError

        index = op_index(index)
        if index < 0:
            index += len(self)

        if 0 <= index < len(self):
            return self._ctx.get_object_item(self, index)

        raise IndexError

    def __setitem__(self, index: int | slice, val: Any) -> None:  # noqa: ANN401
        if not isinstance(index, int):
            raise TypeError

        self._ctx.set_object_item(self, index, val)

    def __delitem__(self, index: int | slice) -> None:
        if not isinstance(index, int):
            raise TypeError

        if index >= len(self) or index < -len(self):
            # JavaScript Array.prototype.splice() just ignores deletion beyond the
            # end of the array, meaning if you pass a very large value here it would
            # do nothing. Likewise, it just caps negative values at the length of the
            # array, meaning if you pass a very negative value here it would just
            # delete element 0.
            # For consistency with Python lists, let's tell the caller they're out of
            # bounds:
            raise JSArrayIndexError

        self._ctx.del_from_array(self, index)

    def insert(self, index: int, new_obj: PythonJSConvertedTypes) -> None:
        self._ctx.array_insert(self, index, new_obj)

    def __iter__(self) -> Iterator[PythonJSConvertedTypes]:
        for i in range(len(self)):
            yield self[i]


class JSFunctionImpl(JSMappedObjectImpl, JSFunction):
    """JavaScript function.

    You can call this object from Python, passing in positional args to match what the
    JavaScript function expects, along with a keyword argument, `timeout_sec`.
    """

    def __call__(
        self,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        return self._ctx.call_function(self, *args, this=this, timeout_sec=timeout_sec)


class JSSymbolImpl(JSMappedObjectImpl, JSSymbol):
    """JavaScript symbol."""


class JSPromiseImpl(JSObjectImpl, JSPromise):
    """JavaScript Promise.

    To get a value, call `promise.get()` to block, or `await promise` from within an
    `async` coroutine. Either will raise a Python exception if the JavaScript Promise
    is rejected.
    """

    def get(self, *, timeout: float | None = None) -> PythonJSConvertedTypes:
        """Get the value, or raise an exception. This call blocks.

        Args:
            timeout: number of milliseconds after which the execution is interrupted.
                This is deprecated; use timeout_sec instead.
        """

        future = SyncFuture()

        def future_caller(value: Any) -> None:  # noqa: ANN401
            future.set_result(value)

        self._attach_callbacks_to_promise(future_caller)

        results = future.get(timeout=timeout)
        return self._unpack_promise_results(results)

    def __await__(self) -> Generator[Any, None, Any]:
        return self._do_await().__await__()

    async def _do_await(self) -> PythonJSConvertedTypes:
        loop = get_running_loop()
        future: Future[PythonJSConvertedTypes] = loop.create_future()

        def future_caller(value: Any) -> None:  # noqa: ANN401
            loop.call_soon_threadsafe(future.set_result, value)

        self._attach_callbacks_to_promise(future_caller)

        results = await future
        return self._unpack_promise_results(results)

    def _attach_callbacks_to_promise(
        self,
        future_caller: Callable[[Any], None],
    ) -> None:
        """Attach the given Python callbacks to a JS Promise."""

        exit_stack = ExitStack()

        def on_resolved_and_cleanup(
            value: PythonJSConvertedTypes | JSEvalException,
        ) -> None:
            exit_stack.__exit__(None, None, None)
            future_caller([False, cast("JSArray", value)])

        on_resolved_js_func = exit_stack.enter_context(
            self._ctx.js_callback(on_resolved_and_cleanup),
        )

        def on_rejected_and_cleanup(
            value: PythonJSConvertedTypes | JSEvalException,
        ) -> None:
            exit_stack.__exit__(None, None, None)
            future_caller([True, cast("JSArray", value)])

        on_rejected_js_func = exit_stack.enter_context(
            self._ctx.js_callback(on_rejected_and_cleanup),
        )

        self._ctx.promise_then(self, on_resolved_js_func, on_rejected_js_func)

    def _unpack_promise_results(self, results: Any) -> PythonJSConvertedTypes:  # noqa: ANN401
        is_rejected, argv = results
        result = cast("JSArray", argv)[0]
        if is_rejected:
            msg = _get_exception_msg(result)
            raise JSPromiseError(msg)
        return result



"""Basic JS types tests"""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from json import dumps
from time import time
from typing import Any, cast

import pytest

from py_mini_racer import (
    JSArray,
    JSEvalException,
    JSFunction,
    JSObject,
    JSSymbol,
    JSUndefined,
    MiniRacer,
)
from tests.gc_check import assert_no_v8_objects


def _test_round_trip(mr: MiniRacer, val: Any) -> None:  # noqa: ANN401
    a = cast("JSArray", mr.eval("[]"))
    a.append(val)  # force conversion into a JS type
    assert a[0] == val  # get it back again and verify it


@dataclass(frozen=True)
class Validator:
    round_trip: bool = True
    mr: MiniRacer = field(default_factory=MiniRacer)

    def __call__(self, py_val: Any) -> None:  # noqa: ANN401
        testee = py_val
        js_str = dumps(py_val)

        parsed = self.mr.execute(js_str)
        assert testee == parsed

        if self.round_trip:
            _test_round_trip(self.mr, py_val)

        assert_no_v8_objects(self.mr)


def test_undefined() -> None:
    mr = MiniRacer()
    undef = mr.eval("undefined")
    assert undef is JSUndefined
    assert undef == JSUndefined
    assert not undef
    _test_round_trip(mr, undef)

    del undef
    assert_no_v8_objects(mr)


def test_str() -> None:
    v = Validator()
    v("'a string'")
    v("'a ' + 'string'")
    v("string with null \0 byte")


def test_unicode() -> None:
    ustr = "\N{GREEK CAPITAL LETTER DELTA}"
    mr = MiniRacer()
    res = mr.eval("'" + ustr + "'")
    assert ustr == res
    _test_round_trip(mr, ustr)

    assert_no_v8_objects(mr)


def test_numbers() -> None:
    v = Validator()
    v(1)
    v(1.0)
    v(2**16)
    v(2**31 - 1)
    v(2**31)
    v(2**33)


def test_arrays() -> None:
    v = Validator(round_trip=False)
    v([1])
    v([])
    v([1, 2, 3])
    # Nested
    v([1, 2, ["a", 1]])


def test_none() -> None:
    v = Validator()
    v(None)


def test_hash() -> None:
    v = Validator(round_trip=False)
    v({})
    v("{}")
    v({"a": 1})
    v({" ": {"z": "www"}})


def test_complex() -> None:
    v = Validator(round_trip=False)
    v(
        {
            "1": [
                1,
                2,
                "qwe",
                {"z": [4, 5, 6, {"eqewr": 1, "zxczxc": "qweqwe", "z": {"1": 2}}]},
            ],
            "qwe": 1,
        },
    )


def test_object() -> None:
    mr = MiniRacer()
    res = mr.eval("var a = {}; a")
    assert isinstance(res, JSObject)
    assert res.__hash__() is not None
    _test_round_trip(mr, res)

    del res
    assert_no_v8_objects(mr)


def test_timestamp() -> None:
    val = int(time())
    mr = MiniRacer()
    res = mr.eval(f"var a = new Date({val * 1000}); a")
    assert res == datetime.fromtimestamp(val, timezone.utc)
    _test_round_trip(mr, res)

    assert_no_v8_objects(mr)


def test_symbol() -> None:
    mr = MiniRacer()
    res = mr.eval('Symbol("my_symbol")')
    assert isinstance(res, JSSymbol)
    assert res.__hash__() is not None
    _test_round_trip(mr, res)

    del res
    assert_no_v8_objects(mr)


def test_function() -> None:
    mr = MiniRacer()
    res = mr.eval("function func() {}; func")
    assert isinstance(res, JSFunction)
    assert res.__hash__() is not None
    _test_round_trip(mr, res)

    del res
    assert_no_v8_objects(mr)


def test_date() -> None:
    mr = MiniRacer()
    res = mr.eval("var a = new Date(Date.UTC(2014, 0, 2, 3, 4, 5)); a")
    assert res == datetime(2014, 1, 2, 3, 4, 5, tzinfo=timezone.utc)
    _test_round_trip(mr, res)

    del res
    assert_no_v8_objects(mr)


def test_exception() -> None:
    js_source = """
    var f = function(arg) {
        throw 'error: '+arg
        return nil
    }"""

    mr = MiniRacer()
    mr.eval(js_source)

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("f(42)")

    assert "error: 42" in exc_info.value.args[0]

    del exc_info
    assert_no_v8_objects(mr)


def test_array_buffer() -> None:
    js_source = """
    var b = new ArrayBuffer(1024);
    var v = new Uint8Array(b);
    v[0] = 0x42;
    b
    """
    mr = MiniRacer()
    ret = cast("memoryview", mr.eval(js_source))
    assert len(ret) == 1024  # noqa: PLR2004
    assert ret[0:1].tobytes() == b"\x42"

    del ret
    assert_no_v8_objects(mr)


def test_array_buffer_view() -> None:
    js_source = """
    var b = new ArrayBuffer(1024);
    var v = new Uint8Array(b, 1, 1);
    v[0] = 0x42;
    v
    """
    mr = MiniRacer()
    ret = cast("memoryview", mr.eval(js_source))
    assert len(ret) == 1
    assert ret.tobytes() == b"\x42"

    del ret
    assert_no_v8_objects(mr)


def test_shared_array_buffer() -> None:
    js_source = """
    var b = new SharedArrayBuffer(1024);
    var v = new Uint8Array(b);
    v[0] = 0x42;
    b
    """
    mr = MiniRacer()
    ret = cast("memoryview", mr.eval(js_source))
    assert len(ret) == 1024  # noqa: PLR2004
    assert ret[0:1].tobytes() == b"\x42"
    ret[1:2] = b"\xff"
    assert mr.eval("v[1]") == 0xFF  # noqa: PLR2004

    del ret
    assert_no_v8_objects(mr)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest

from py_mini_racer import (
    JSArray,
    JSFunction,
    JSObject,
    JSPromise,
    JSSymbol,
    JSUndefined,
    MiniRacer,
)
from tests.gc_check import assert_no_v8_objects

if TYPE_CHECKING:
    from py_mini_racer import JSMappedObject


def test_object_read() -> None:
    mr = MiniRacer()
    obj = cast(
        "JSMappedObject",
        mr.eval(
            """\
var a = {
    5: "key_is_number",
    "key_is_string": 42,
    undefined: "undef_value",
    null: "null_value",
};
a
""",
        ),
    )

    assert obj.__hash__()
    assert obj
    assert sorted(obj.keys(), key=str) == [5, "key_is_string", "null", "undefined"]
    assert obj["5"] == "key_is_number"
    assert obj[5] == "key_is_number"
    assert obj["key_is_string"] == 42  # noqa: PLR2004
    assert obj[None] == "null_value"
    assert obj[JSUndefined] == "undef_value"
    assert len(obj) == 4  # noqa: PLR2004

    # The following are provided by collections.abc mixins:
    assert 5 in obj  # noqa: PLR2004
    assert None in obj
    assert "elvis" not in obj
    assert sorted(obj.items(), key=lambda x: str(x[0])) == [
        (5, "key_is_number"),
        ("key_is_string", 42),
        ("null", "null_value"),
        ("undefined", "undef_value"),
    ]
    assert set(obj.values()) == {42, "key_is_number", "undef_value", "null_value"}
    obj2 = mr.eval(
        """\
var a = {
    5: "key_is_number",
    "key_is_string": 42,
    undefined: "undef_value",
    null: "null_value",
};
a
""",
    )
    assert obj == obj2

    obj3 = mr.eval(
        """\
var a = {};
a
""",
    )
    assert not obj3

    del obj, obj2, obj3
    assert_no_v8_objects(mr)


def test_object_mutation() -> None:
    mr = MiniRacer()
    obj = cast(
        "JSMappedObject",
        mr.eval(
            """\
var a = {};
a
""",
        ),
    )

    obj["some_string"] = "some_string_val"
    obj[JSUndefined] = "undefined_val"
    obj[None] = "none_val"
    obj[5] = "int_val"
    # Note that this should overwrite 5=int_val above:
    obj[5.0] = "double_val"
    assert sorted(obj.items(), key=lambda x: str(x[0])) == [
        (5, "double_val"),
        ("null", "none_val"),
        ("some_string", "some_string_val"),
        ("undefined", "undefined_val"),
    ]
    assert obj.pop(None) == "none_val"
    with pytest.raises(KeyError):
        obj.pop(None)
    with pytest.raises(KeyError):
        del obj["elvis"]
    obj.clear()
    assert not obj
    obj["foo"] = "bar"
    assert obj.setdefault("foo", "baz") == "bar"
    obj.update({"froz": "blargh"})
    assert len(obj) == 2  # noqa: PLR2004

    inner_obj = mr.eval(
        """\
var b = {"k": "v"};
b
""",
    )
    obj["inner"] = inner_obj
    assert len(obj) == 3  # noqa: PLR2004
    assert cast("JSMappedObject", obj["inner"])["k"] == "v"

    del obj, inner_obj
    assert_no_v8_objects(mr)


def test_object_prototype() -> None:
    mr = MiniRacer()
    obj = cast(
        "JSMappedObject",
        mr.eval(
            """\
var proto = { 5: "key_is_number", "key_is_string": 42 };
var a = Object.create(proto);
a.foo = "bar";
a
""",
        ),
    )
    assert sorted(obj.items(), key=lambda x: str(x[0])) == [
        (5, "key_is_number"),
        ("foo", "bar"),
        ("key_is_string", 42),
    ]

    del obj
    assert_no_v8_objects(mr)


def test_array() -> None:
    mr = MiniRacer()
    obj = mr.eval(
        """\
var a = [ "some_string", 42, undefined, null ];
a
""",
    )

    assert isinstance(obj, JSArray)
    assert obj.__hash__()
    assert obj
    assert obj[0] == "some_string"
    assert obj[1] == 42  # noqa: PLR2004
    assert obj[2] is JSUndefined
    assert obj[2] == JSUndefined
    assert obj[3] is None
    assert obj[-3] == 42  # noqa: PLR2004
    with pytest.raises(IndexError):
        obj[4]
    with pytest.raises(IndexError):
        obj[-5]

    assert list(obj) == ["some_string", 42, JSUndefined, None]
    assert len(obj) == 4  # noqa: PLR2004
    assert list(obj) == ["some_string", 42, JSUndefined, None]
    assert 42 in obj  # noqa: PLR2004
    assert JSUndefined in obj
    assert None in obj
    assert "elvis" not in obj

    obj2 = mr.eval(
        """\
var a = [];
a
""",
    )
    assert not obj2

    del obj, obj2
    assert_no_v8_objects(mr)


def test_array_mutation() -> None:
    mr = MiniRacer()
    obj = cast(
        "JSArray",
        mr.eval(
            """\
var a = [];
a
""",
        ),
    )

    obj.append("some_string")
    obj.append(JSUndefined)
    obj.insert(1, 42)
    obj.insert(-1, None)
    assert list(obj) == ["some_string", 42, None, JSUndefined]

    del obj[-1]
    assert list(obj) == ["some_string", 42, None]

    del obj[0]
    assert list(obj) == [42, None]

    with pytest.raises(IndexError):
        del obj[-3]

    with pytest.raises(IndexError):
        del obj[2]

    inner_obj = mr.eval(
        """\
var b = {"k": "v"};
b
""",
    )
    obj.append(inner_obj)
    assert len(obj) == 3  # noqa: PLR2004
    assert cast("JSMappedObject", obj[-1])["k"] == "v"

    del obj, inner_obj
    assert_no_v8_objects(mr)


def test_function() -> None:
    mr = MiniRacer()
    obj = mr.eval(
        """\
function foo() {};
foo
""",
    )

    assert isinstance(obj, JSFunction)
    assert obj.__hash__()
    assert tuple(obj.keys()) == ()

    del obj
    assert_no_v8_objects(mr)


def test_symbol() -> None:
    mr = MiniRacer()
    obj = mr.eval(
        """\
var sym = Symbol("foo");
sym
""",
    )

    assert isinstance(obj, JSSymbol)
    assert obj.__hash__()
    assert tuple(obj.keys()) == ()

    del obj
    assert_no_v8_objects(mr)


def test_promise() -> None:
    mr = MiniRacer()
    promise = mr.eval(
        """\
var p = Promise.resolve(42);
p
""",
    )

    assert isinstance(promise, JSPromise)
    assert promise.__hash__()

    del promise
    assert_no_v8_objects(mr)


def test_nested_object() -> None:
    mr = MiniRacer()
    obj = cast(
        "JSMappedObject",
        mr.eval(
            """\
var a = {
    5: "key_is_number",
    "key_is_string": 42,
    "some_func": () => {},
    "some_obj": {"a": 12},
    "some_promise": Promise.resolve(42),
    "some_symbol": Symbol("sym"),
};
a
""",
        ),
    )

    assert obj.__hash__()
    assert sorted(obj.keys(), key=str) == [
        5,
        "key_is_string",
        "some_func",
        "some_obj",
        "some_promise",
        "some_symbol",
    ]
    assert obj["5"] == "key_is_number"
    assert obj[5] == "key_is_number"
    assert obj["key_is_string"] == 42  # noqa: PLR2004
    assert isinstance(obj["some_func"], JSFunction)
    assert isinstance(obj["some_obj"], JSObject)
    assert cast("JSMappedObject", obj["some_obj"])["a"] == 12  # noqa: PLR2004
    assert isinstance(obj["some_promise"], JSPromise)
    assert isinstance(obj["some_symbol"], JSSymbol)

    del obj
    assert_no_v8_objects(mr)



from __future__ import annotations

import ctypes
from datetime import datetime, timezone
from typing import (
    TYPE_CHECKING,
    ClassVar,
)

from py_mini_racer._abstract_context import AbstractContext, AbstractValueHandle
from py_mini_racer._exc import JSEvalException, MiniRacerBaseException
from py_mini_racer._objects import (
    JSArrayImpl,
    JSFunctionImpl,
    JSMappedObjectImpl,
    JSObjectImpl,
    JSPromiseImpl,
    JSSymbolImpl,
)
from py_mini_racer._types import (
    JSUndefined,
    PythonJSConvertedTypes,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


class _RawValueUnion(ctypes.Union):
    _fields_: ClassVar[Sequence[tuple[str, type]]] = [
        ("value_ptr", ctypes.c_void_p),
        ("bytes_val", ctypes.POINTER(ctypes.c_char)),
        ("char_p_val", ctypes.c_char_p),
        ("int_val", ctypes.c_int64),
        ("double_val", ctypes.c_double),
    ]


class _RawValue(ctypes.Structure):
    _fields_: ClassVar[Sequence[tuple[str, type]]] = [
        ("value", _RawValueUnion),
        ("len", ctypes.c_size_t),
        ("type", ctypes.c_uint8),
    ]
    _pack_ = 1


RawValueHandle = ctypes.POINTER(_RawValue)

if TYPE_CHECKING:
    RawValueHandleType = ctypes._Pointer[_RawValue]  # noqa: SLF001


class _ArrayBufferByte(ctypes.Structure):
    # Cannot use c_ubyte directly because it uses <B
    # as an internal type but we need B for memoryview.
    _fields_: ClassVar[Sequence[tuple[str, type]]] = [("b", ctypes.c_ubyte)]
    _pack_ = 1


class JSParseException(JSEvalException):
    """JavaScript could not be parsed."""


class JSKeyError(JSEvalException, KeyError):
    """No such key found."""


class JSOOMException(JSEvalException):
    """JavaScript execution ran out of memory."""


class JSTerminatedException(JSEvalException):
    """JavaScript execution terminated."""


class JSValueError(JSEvalException, ValueError):
    """Bad value passed to JavaScript engine."""


class JSConversionException(MiniRacerBaseException):
    """Type could not be converted to or from JavaScript."""


class MiniRacerTypes:
    """MiniRacer types identifier

    Note: it needs to be coherent with mini_racer.cc.
    """

    invalid = 0
    null = 1
    bool = 2
    integer = 3
    double = 4
    str_utf8 = 5
    array = 6
    # deprecated:
    hash = 7
    date = 8
    symbol = 9
    object = 10
    undefined = 11

    function = 100
    shared_array_buffer = 101
    array_buffer = 102
    promise = 103

    execute_exception = 200
    parse_exception = 201
    oom_exception = 202
    timeout_exception = 203
    terminated_exception = 204
    value_exception = 205
    key_exception = 206


_ERRORS: dict[int, tuple[type[JSEvalException], str]] = {
    MiniRacerTypes.parse_exception: (
        JSParseException,
        "Unknown JavaScript error during parse",
    ),
    MiniRacerTypes.execute_exception: (
        JSEvalException,
        "Uknown JavaScript error during execution",
    ),
    MiniRacerTypes.oom_exception: (JSOOMException, "JavaScript memory limit reached"),
    MiniRacerTypes.terminated_exception: (
        JSTerminatedException,
        "JavaScript was terminated",
    ),
    MiniRacerTypes.key_exception: (
        JSKeyError,
        "No such key found in object",
    ),
    MiniRacerTypes.value_exception: (
        JSValueError,
        "Bad value passed to JavaScript engine",
    ),
}


class ValueHandle(AbstractValueHandle):
    """An object which holds open a Python reference to a _RawValue owned by
    a C++ MiniRacer context."""

    def __init__(self, ctx: AbstractContext, raw: RawValueHandleType) -> None:
        self.ctx = ctx
        self._raw = raw

    def __del__(self) -> None:
        self.ctx.free(self)

    @property
    def raw(self) -> RawValueHandleType:
        return self._raw

    def to_python_or_raise(self) -> PythonJSConvertedTypes:
        val = self.to_python()
        if isinstance(val, JSEvalException):
            raise val
        return val

    def to_python(self) -> PythonJSConvertedTypes | JSEvalException:  # noqa: C901, PLR0911, PLR0912
        """Convert a binary value handle from the C++ side into a Python object."""

        # A MiniRacer binary value handle is a pointer to a structure which, for some
        # simple types like ints, floats, and strings, is sufficient to describe the
        # data, enabling us to convert the value immediately and free the handle.

        # For more complex types, like Objects and Arrays, the handle is just an opaque
        # pointer to a V8 object. In these cases, we retain the binary value handle,
        # wrapping it in a Python object. We can then use the handle in follow-on API
        # calls to work with the underlying V8 object.

        # In either case the handle is owned by the C++ side. It's the responsibility
        # of the Python side to call mr_free_value() when done with with the handle
        # to free up memory, but the C++ side will eventually free it on context
        # teardown either way.

        typ = self._raw.contents.type
        val = self._raw.contents.value
        length = self._raw.contents.len

        error_info = _ERRORS.get(self._raw.contents.type)
        if error_info:
            klass, generic_msg = error_info

            msg = val.bytes_val[0:length].decode("utf-8")
            msg = msg or generic_msg

            return klass(msg)

        if typ == MiniRacerTypes.null:
            return None
        if typ == MiniRacerTypes.undefined:
            return JSUndefined
        if typ == MiniRacerTypes.bool:
            return bool(val.int_val == 1)
        if typ == MiniRacerTypes.integer:
            return int(val.int_val)
        if typ == MiniRacerTypes.double:
            return float(val.double_val)
        if typ == MiniRacerTypes.str_utf8:
            return str(val.bytes_val[0:length].decode("utf-8"))
        if typ == MiniRacerTypes.function:
            return JSFunctionImpl(self.ctx, self)
        if typ == MiniRacerTypes.date:
            timestamp = val.double_val
            # JS timestamps are milliseconds. In Python we are in seconds:
            return datetime.fromtimestamp(timestamp / 1000.0, timezone.utc)
        if typ == MiniRacerTypes.symbol:
            return JSSymbolImpl(self.ctx, self)
        if typ in (MiniRacerTypes.shared_array_buffer, MiniRacerTypes.array_buffer):
            buf = _ArrayBufferByte * length
            cdata = buf.from_address(val.value_ptr)
            # Save a reference to ourselves to prevent garbage collection of the
            # backing store:
            cdata._origin = self  # noqa: SLF001
            result = memoryview(cdata)
            # Avoids "NotImplementedError: memoryview: unsupported format T{<B:b:}"
            # in Python 3.12:
            return result.cast("B")

        if typ == MiniRacerTypes.promise:
            return JSPromiseImpl(self.ctx, self)

        if typ == MiniRacerTypes.array:
            return JSArrayImpl(self.ctx, self)

        if typ == MiniRacerTypes.object:
            return JSMappedObjectImpl(self.ctx, self)

        raise JSConversionException


def python_to_value_handle(  # noqa: PLR0911
    context: AbstractContext, obj: PythonJSConvertedTypes
) -> AbstractValueHandle:
    if isinstance(obj, JSObjectImpl):
        # JSObjects originate from the V8 side. We can just send back the handle
        # we originally got. (This also covers derived types JSFunction, JSSymbol,
        # JSPromise, and JSArray.)
        return obj.raw_handle

    if obj is None:
        return context.create_intish_val(0, MiniRacerTypes.null)
    if obj is JSUndefined:
        return context.create_intish_val(0, MiniRacerTypes.undefined)
    if isinstance(obj, bool):
        return context.create_intish_val(1 if obj else 0, MiniRacerTypes.bool)
    if isinstance(obj, int):
        if obj - 2**31 <= obj < 2**31:
            return context.create_intish_val(obj, MiniRacerTypes.integer)

        # We transmit ints as int32, so "upgrade" to double upon overflow.
        # (ECMAScript numeric is double anyway, but V8 does internally distinguish
        # int types, so we try and preserve integer-ness for round-tripping
        # purposes.)
        # JS BigInt would be a closer representation of Python int, but upgrading
        # to BigInt would probably be surprising for most applications, so for now,
        # we approximate with double:
        return context.create_doublish_val(obj, MiniRacerTypes.double)
    if isinstance(obj, float):
        return context.create_doublish_val(obj, MiniRacerTypes.double)
    if isinstance(obj, str):
        return context.create_string_val(obj, MiniRacerTypes.str_utf8)
    if isinstance(obj, datetime):
        # JS timestamps are milliseconds. In Python we are in seconds:
        return context.create_doublish_val(
            obj.timestamp() * 1000.0,
            MiniRacerTypes.date,
        )

    # Note: we skip shared array buffers, so for now at least, handles to shared
    # array buffers can only be transmitted from JS to Python.

    raise JSConversionException



from __future__ import annotations

from py_mini_racer._context import PyJsFunctionType
from py_mini_racer._dll import (
    DEFAULT_V8_FLAGS,
    LibAlreadyInitializedError,
    LibNotFoundError,
    init_mini_racer,
)
from py_mini_racer._exc import (
    JSArrayIndexError,
    JSEvalException,
    JSPromiseError,
    JSTimeoutException,
)
from py_mini_racer._mini_racer import MiniRacer, StrictMiniRacer
from py_mini_racer._types import (
    JSArray,
    JSFunction,
    JSMappedObject,
    JSObject,
    JSPromise,
    JSSymbol,
    JSUndefined,
    JSUndefinedType,
    PythonJSConvertedTypes,
)
from py_mini_racer._value_handle import (
    JSKeyError,
    JSOOMException,
    JSParseException,
    JSValueError,
)

__all__ = [
    "DEFAULT_V8_FLAGS",
    "AsyncCleanupType",
    "JSArray",
    "JSArrayIndexError",
    "JSEvalException",
    "JSFunction",
    "JSKeyError",
    "JSMappedObject",
    "JSOOMException",
    "JSObject",
    "JSParseException",
    "JSPromise",
    "JSPromiseError",
    "JSSymbol",
    "JSTimeoutException",
    "JSUndefined",
    "JSUndefinedType",
    "JSValueError",
    "LibAlreadyInitializedError",
    "LibNotFoundError",
    "MiniRacer",
    "PyJsFunctionType",
    "PythonJSConvertedTypes",
    "StrictMiniRacer",
    "init_mini_racer",
]



"""Test .eval() method"""

from __future__ import annotations

from asyncio import run as asyncio_run
from time import sleep, time
from typing import cast

import pytest

from py_mini_racer import (
    JSEvalException,
    JSOOMException,
    JSParseException,
    JSPromise,
    JSPromiseError,
    JSSymbol,
    JSTimeoutException,
    JSUndefined,
    MiniRacer,
)
from tests.gc_check import assert_no_v8_objects

# Wait time for async tests to finish.
_ASYNC_COMPLETION_WAIT_SEC = 10


def test_invalid() -> None:
    mr = MiniRacer()

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("invalid")

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: ReferenceError: invalid is not defined
invalid
^

ReferenceError: invalid is not defined
    at <anonymous>:1:1
"""
    )

    del exc_info
    assert_no_v8_objects(mr)


def test_eval() -> None:
    mr = MiniRacer()
    assert mr.eval("42") == 42  # noqa: PLR2004

    assert_no_v8_objects(mr)


def test_blank() -> None:
    mr = MiniRacer()
    assert mr.eval("") is JSUndefined
    assert mr.eval(" ") is JSUndefined
    assert mr.eval("\t") is JSUndefined

    assert_no_v8_objects(mr)


def test_global() -> None:
    mr = MiniRacer()
    mr.eval("var xabc = 22;")
    assert mr.eval("xabc") == 22  # noqa: PLR2004

    assert_no_v8_objects(mr)


def test_fun() -> None:
    mr = MiniRacer()
    mr.eval("var x = function(y) {return y+1;}")

    assert mr.eval("x(1)") == 2  # noqa: PLR2004
    assert mr.eval("x(10)") == 11  # noqa: PLR2004
    assert mr.eval("x(100)") == 101  # noqa: PLR2004

    assert_no_v8_objects(mr)


def test_multiple_ctx() -> None:
    c1 = MiniRacer()
    c2 = MiniRacer()
    c3 = MiniRacer()

    c1.eval("var x = 1")
    c2.eval("var x = 2")
    c3.eval("var x = 3")
    assert c1.eval("(x)") == 1
    assert c2.eval("(x)") == 2  # noqa: PLR2004
    assert c3.eval("(x)") == 3  # noqa: PLR2004

    assert_no_v8_objects(c1)
    assert_no_v8_objects(c2)
    assert_no_v8_objects(c3)


def test_exception_thrown() -> None:
    mr = MiniRacer()

    js_source = "var f = function() {throw new Error('blah')};"

    mr.eval(js_source)

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("f()")

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: Error: blah
var f = function() {throw new Error('blah')};
                    ^

Error: blah
    at f (<anonymous>:1:27)
    at <anonymous>:1:1
"""
    )

    del exc_info
    assert_no_v8_objects(mr)


def test_string_thrown() -> None:
    mr = MiniRacer()

    js_source = "var f = function() {throw 'blah'};"

    mr.eval(js_source)

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("f()")

    # When you throw a plain string (not wrapping it in a `new Error(...)`), you
    # get no backtrace:
    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: blah
var f = function() {throw 'blah'};
                    ^
"""
    )

    del exc_info
    assert_no_v8_objects(mr)


def test_cannot_parse() -> None:
    mr = MiniRacer()
    js_source = "var f = function("

    with pytest.raises(JSParseException) as exc_info:
        mr.eval(js_source)

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: SyntaxError: Unexpected end of input
var f = function(
                 ^

SyntaxError: Unexpected end of input
"""
    )

    del exc_info
    assert_no_v8_objects(mr)


def test_null_byte() -> None:
    mr = MiniRacer()

    s = "\x00 my string!"

    # Try return a string including a null byte
    in_val = 'var str = "' + s + '"; str;'
    result = mr.eval(in_val)
    assert result == s

    assert_no_v8_objects(mr)


def test_timeout() -> None:
    timeout = 0.3
    start_time = time()

    mr = MiniRacer()
    with pytest.raises(JSTimeoutException) as exc_info:
        mr.eval("while(1) { }", timeout_sec=timeout)

    duration = time() - start_time
    # Make sure it timed out on time, and allow a large leeway
    assert timeout * 0.9 <= duration <= timeout + 2

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"

    del exc_info
    assert_no_v8_objects(mr)


def test_timeout_ms() -> None:
    # Same as above but with the deprecated timeout millisecond argument
    timeout = 0.3
    start_time = time()

    mr = MiniRacer()
    with pytest.raises(JSTimeoutException) as exc_info:
        mr.eval("while(1) { }", timeout=int(timeout * 1000))

    duration = time() - start_time
    # Make sure it timed out on time, and allow a large leeway
    assert timeout * 0.9 <= duration <= timeout + 2

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"

    del exc_info
    assert_no_v8_objects(mr)


def test_max_memory_soft() -> None:
    mr = MiniRacer()
    mr.set_soft_memory_limit(100000000)
    mr.set_hard_memory_limit(100000000)
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """\
let s = 1000;
var a = new Array(s);
a.fill(0);
while(true) {
    s *= 1.1;
    let n = new Array(Math.floor(s));
    n.fill(0);
    a = a.concat(n);
}
""",
        )

    assert mr.was_soft_memory_limit_reached()
    assert mr.was_hard_memory_limit_reached()
    assert exc_info.value.args[0] == "JavaScript memory limit reached"

    del exc_info
    assert_no_v8_objects(mr)


def test_max_memory_hard() -> None:
    mr = MiniRacer()
    mr.set_hard_memory_limit(100000000)
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """\
let s = 1000;
var a = new Array(s);
a.fill(0);
while(true) {
    s *= 1.1;
    let n = new Array(Math.floor(s));
    n.fill(0);
    a = a.concat(n);
}""",
        )

    assert not mr.was_soft_memory_limit_reached()
    assert mr.was_hard_memory_limit_reached()
    assert exc_info.value.args[0] == "JavaScript memory limit reached"

    del exc_info
    assert_no_v8_objects(mr)


def test_max_memory_hard_eval_arg() -> None:
    # Same as above but passing the argument into the eval method (which is a
    # deprecated thing to do because the parameter is really affine to the
    # MiniRacer object)
    mr = MiniRacer()
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """\
let s = 1000;
var a = new Array(s);
a.fill(0);
while(true) {
    s *= 1.1;
    let n = new Array(Math.floor(s));
    n.fill(0);
    a = a.concat(n);
}""",
            max_memory=200000000,
        )

    assert exc_info.value.args[0] == "JavaScript memory limit reached"

    del exc_info
    assert_no_v8_objects(mr)


def test_symbol() -> None:
    mr = MiniRacer()
    res = mr.eval("Symbol.toPrimitive")
    assert isinstance(res, JSSymbol)

    del res
    assert_no_v8_objects(mr)


def test_microtask() -> None:
    # PyMiniRacer uses V8 microtasks (things, like certain promise callbacks, which run
    # immediately after an evaluation ends).
    # By default, V8 runs any microtasks before it returns control to PyMiniRacer.
    # Let's test that they actually work.
    # PyMiniRacer does not expose the web standard `window.queueMicrotask` (because it
    # does not expose a `window` to begin with). We can, however, trigger a microtask
    # by triggering one as a side effect of a `then` on a resolved promise:
    mr = MiniRacer()
    assert not mr.eval(
        """
let p = Promise.resolve();

var done = false;

p.then(() => {done = true});

done
""",
    )
    assert mr.eval("done")

    assert_no_v8_objects(mr)


def test_longer_microtask() -> None:
    # Verifies a bug fix wherein failure to set a v8::Isolate::Scope on the message
    # pump thread would otherwise result in a segmentation fault:
    mr = MiniRacer()
    mr.eval(
        """
var done = false;
async function foo() {
    await new Promise((res, rej) => setTimeout(res, 1000));
    for (let i = 0; i < 10000000; i++) { }
    done = true;
}
foo();
""",
    )

    assert not mr.eval("done")
    start = time()
    while time() - start < _ASYNC_COMPLETION_WAIT_SEC and not mr.eval("done"):
        sleep(0.1)
    assert mr.eval("done")

    assert_no_v8_objects(mr)


def test_polling() -> None:
    mr = MiniRacer()
    assert not mr.eval(
        """
var done = false;
setTimeout(() => { done = true; }, 1000);
done
""",
    )
    assert not mr.eval("done")
    start = time()
    while time() - start < _ASYNC_COMPLETION_WAIT_SEC and not mr.eval("done"):
        sleep(0.1)
    assert mr.eval("done")

    assert_no_v8_objects(mr)


def test_settimeout() -> None:
    mr = MiniRacer()
    mr.eval(
        """
var results = [];
let a = setTimeout(() => { results.push("a"); }, 2000);
let b = setTimeout(() => { results.push("b"); }, 3000);
let c = setTimeout(() => { results.push("c"); }, 1000);
let d = setTimeout(() => { results.push("d"); }, 4000);
clearTimeout(b)
""",
    )
    start = time()
    while (
        time() - start < _ASYNC_COMPLETION_WAIT_SEC and mr.eval("results.length") != 3  # noqa: PLR2004
    ):
        sleep(0.1)
    assert mr.eval("results.length") == 3  # noqa: PLR2004
    assert mr.eval("results[0]") == "c"
    assert mr.eval("results[1]") == "a"
    assert mr.eval("results[2]") == "d"

    assert_no_v8_objects(mr)


def test_promise_sync() -> None:
    mr = MiniRacer()
    p = cast(
        "JSPromise",
        mr.eval(
            """
new Promise((res, rej) => setTimeout(() => res(42), 1000)); // 1 s timeout
""",
        ),
    )
    start = time()
    result = p.get(timeout=10)
    assert time() - start > 0.5  # noqa: PLR2004
    assert result == 42  # noqa: PLR2004

    del p
    assert_no_v8_objects(mr)


def test_promise_async() -> None:
    mr = MiniRacer()

    async def run_test() -> None:
        p = cast(
            "JSPromise",
            mr.eval(
                """
new Promise((res, rej) => setTimeout(() => res(42), 1000)); // 1 s timeout
""",
            ),
        )
        start = time()
        result = await p
        assert time() - start > 0.5  # noqa: PLR2004
        assert time() - start < _ASYNC_COMPLETION_WAIT_SEC
        assert result == 42  # noqa: PLR2004

    asyncio_run(run_test())
    assert_no_v8_objects(mr)


def test_resolved_promise_sync() -> None:
    mr = MiniRacer()
    p = cast("JSPromise", mr.eval("Promise.resolve(6*7)"))
    val = p.get()
    assert val == 42  # noqa: PLR2004

    del val, p

    assert_no_v8_objects(mr)


def test_resolved_promise_async() -> None:
    mr = MiniRacer()

    async def run_test() -> None:
        p = cast("JSPromise", mr.eval("Promise.resolve(6*7)"))
        val = await p
        assert val == 42  # noqa: PLR2004

    asyncio_run(run_test())
    assert_no_v8_objects(mr)


def test_rejected_promise_sync() -> None:
    mr = MiniRacer()

    p = cast("JSPromise", mr.eval("Promise.reject(new Error('this is an error'))"))
    with pytest.raises(JSPromiseError) as exc_info:
        p.get()

    assert (
        exc_info.value.args[0]
        == """\
JavaScript rejected promise with reason: Error: this is an error
    at <anonymous>:1:16
"""
    )

    del exc_info, p
    assert_no_v8_objects(mr)


def test_rejected_promise_async() -> None:
    mr = MiniRacer()

    async def run_test() -> None:
        p = cast("JSPromise", mr.eval("Promise.reject(new Error('this is an error'))"))
        with pytest.raises(JSPromiseError) as exc_info:
            await p

        assert (
            exc_info.value.args[0]
            == """\
JavaScript rejected promise with reason: Error: this is an error
    at <anonymous>:1:16
"""
        )

    asyncio_run(run_test())
    assert_no_v8_objects(mr)


def test_rejected_promise_sync_stringerror() -> None:
    mr = MiniRacer()

    p = cast("JSPromise", mr.eval("Promise.reject('this is a string')"))
    with pytest.raises(JSPromiseError) as exc_info:
        p.get()

    assert (
        exc_info.value.args[0]
        == """\
JavaScript rejected promise with reason: this is a string
"""
    )

    del exc_info, p
    assert_no_v8_objects(mr)



"""Test loading and executing babel.js"""

from __future__ import annotations

from pathlib import Path

from py_mini_racer import MiniRacer
from tests.gc_check import assert_no_v8_objects


def test_babel() -> None:
    mr = MiniRacer()
    source = f"""
      var self = this;
      {(Path(__file__).parent / "fixtures" / "babel.js").read_text(encoding="utf-8")}
      babel.eval = function(code) {{
        return eval(babel.transform(code)["code"]);
      }}
    """

    mr.eval(source)

    assert mr.eval("babel.eval(((x) => x * x)(8))") == 64  # noqa: PLR2004
    assert_no_v8_objects(mr)



from __future__ import annotations

from traceback import clear_frames

import pytest

from py_mini_racer import JSEvalException, JSUndefined, StrictMiniRacer
from tests.gc_check import assert_no_v8_objects


def test_basic_int() -> None:
    mr = StrictMiniRacer()
    assert mr.execute("42") == 42  # noqa: PLR2004

    assert_no_v8_objects(mr)


def test_basic_string() -> None:
    mr = StrictMiniRacer()
    assert mr.execute('"42"') == "42"

    assert_no_v8_objects(mr)


def test_basic_hash() -> None:
    mr = StrictMiniRacer()
    assert mr.execute("{}") == {}

    assert_no_v8_objects(mr)


def test_basic_array() -> None:
    mr = StrictMiniRacer()
    assert mr.execute("[1, 2, 3]") == [1, 2, 3]

    assert_no_v8_objects(mr)


def test_call() -> None:
    js_func = """var f = function(args) {
        return args.length;
    }"""

    mr = StrictMiniRacer()

    assert mr.eval(js_func) is JSUndefined
    assert mr.call("f", list(range(5))) == 5  # noqa: PLR2004

    assert_no_v8_objects(mr)


def test_message() -> None:
    mr = StrictMiniRacer()
    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("throw new EvalError('Hello', 'someFile.js', 10);")

    clear_frames(exc_info.tb)
    assert_no_v8_objects(mr)



"""Test executing a WASM module."""

from __future__ import annotations

from pathlib import Path
from typing import cast

from py_mini_racer import MiniRacer
from tests.gc_check import assert_no_v8_objects

test_dir = Path(__file__).parent


def test_add() -> None:
    fn = test_dir / "add.wasm"
    mr = MiniRacer()

    # 1. Allocate a buffer to hold the WASM module code
    size = Path(fn).stat().st_size
    module_raw = cast(
        "memoryview",
        mr.eval(
            f"""
    const moduleRaw = new SharedArrayBuffer({size});
    moduleRaw
    """,
        ),
    )

    # 2. Read the WASM module code
    with Path(fn).open("rb") as f:
        assert f.readinto(module_raw) == size

    # 3. Instantiate the WASM module
    mr.eval(
        """
    var res = null;
    WebAssembly.instantiate(new Uint8Array(moduleRaw)).then(result => {
        res = result.instance;
    }).catch(result => { res = result.message; });
    """,
    )

    # 4. Wait for WASM module instantiation
    while mr.eval("res") is None:
        pass

    assert mr.eval("typeof res !== 'string'")

    # 5. Execute a WASM function
    assert mr.eval("res.exports.addTwo(1, 2)") == 3  # noqa: PLR2004

    del module_raw
    assert_no_v8_objects(mr)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
)

from py_mini_racer._types import JSUndefined

if TYPE_CHECKING:
    from collections.abc import Callable
    from contextlib import AbstractContextManager

    from py_mini_racer._exc import JSEvalException
    from py_mini_racer._types import (
        JSArray,
        JSFunction,
        JSObject,
        JSPromise,
        JSUndefinedType,
        PythonJSConvertedTypes,
    )


class AbstractValueHandle(ABC):
    @property
    @abstractmethod
    def raw(self) -> object:
        pass

    @abstractmethod
    def to_python(self) -> PythonJSConvertedTypes | JSEvalException:
        pass

    @abstractmethod
    def to_python_or_raise(self) -> PythonJSConvertedTypes:
        pass


class AbstractContext(ABC):
    """A Context provides Pythonic wrappers around the MiniRacer C API.

    This is intended for internal usage by py_mini_racer. MiniRacerContext provides
    a further wrapper around this interface.
    """

    @abstractmethod
    def get_identity_hash(self, obj: JSObject) -> int:
        pass

    @abstractmethod
    def get_own_property_names(
        self,
        obj: JSObject,
    ) -> tuple[PythonJSConvertedTypes, ...]:
        pass

    @abstractmethod
    def get_object_item(
        self,
        obj: JSObject,
        key: PythonJSConvertedTypes,
    ) -> PythonJSConvertedTypes:
        pass

    @abstractmethod
    def set_object_item(
        self,
        obj: JSObject,
        key: PythonJSConvertedTypes,
        val: PythonJSConvertedTypes,
    ) -> None:
        pass

    @abstractmethod
    def del_object_item(self, obj: JSObject, key: PythonJSConvertedTypes) -> None:
        pass

    @abstractmethod
    def del_from_array(self, arr: JSArray, index: int) -> None:
        pass

    @abstractmethod
    def array_insert(
        self,
        arr: JSArray,
        index: int,
        new_val: PythonJSConvertedTypes,
    ) -> None:
        pass

    @abstractmethod
    def call_function(
        self,
        func: JSFunction,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        pass

    @abstractmethod
    def js_callback(
        self,
        func: Callable[[PythonJSConvertedTypes | JSEvalException], None],
    ) -> AbstractContextManager[JSFunction]:
        pass

    @abstractmethod
    def promise_then(
        self,
        promise: JSPromise,
        on_resolved: JSFunction,
        on_rejected: JSFunction,
    ) -> None:
        pass

    @abstractmethod
    def create_intish_val(self, val: int, typ: int) -> AbstractValueHandle:
        pass

    @abstractmethod
    def create_doublish_val(self, val: float, typ: int) -> AbstractValueHandle:
        pass

    @abstractmethod
    def create_string_val(self, val: str, typ: int) -> AbstractValueHandle:
        pass

    @abstractmethod
    def free(self, val_handle: AbstractValueHandle) -> None:
        pass

    @abstractmethod
    def evaluate(
        self,
        code: str,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        pass



"""Basic JS call functions"""

from __future__ import annotations

from datetime import datetime, timezone
from json import JSONEncoder
from typing import Any

from py_mini_racer import MiniRacer
from tests.gc_check import assert_no_v8_objects


def test_call_js() -> None:
    js_func = """var f = function() {
        return arguments.length;
    }"""

    mr = MiniRacer()
    mr.eval(js_func)

    assert mr.call("f") == 0
    assert mr.call("f", *list(range(5))) == 5  # noqa: PLR2004
    assert mr.call("f", *list(range(10))) == 10  # noqa: PLR2004
    assert mr.call("f", *list(range(20))) == 20  # noqa: PLR2004

    assert_no_v8_objects(mr)


def test_call_custom_encoder() -> None:
    # Custom encoder for dates
    class CustomEncoder(JSONEncoder):
        def default(self, obj: Any) -> str:  # noqa: ANN401
            if isinstance(obj, datetime):
                return obj.isoformat()

            return str(JSONEncoder.default(self, obj))

    now = datetime.now(tz=timezone.utc)
    js_func = """var f = function(args) {
        return args;
    }"""
    mr = MiniRacer()
    mr.eval(js_func)

    assert mr.call("f", now, encoder=CustomEncoder) == now.isoformat()

    assert_no_v8_objects(mr)



from __future__ import annotations

from pathlib import Path

from packaging.tags import platform_tags
from setuptools import setup
from wheel.bdist_wheel import bdist_wheel  # type: ignore[import-untyped]


def _get_platform_tag() -> str:
    """Return a pip platform tag indicating compatibility of the mini_racer binary.

    See https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/.
    """

    # generally return the first, meaning the most-specific, platform tag:
    tag = next(platform_tags())

    if tag.startswith("macos_"):
        # pip seems finicky about platform tags with larger macos versions, so just
        # tell arm64 is 11.0 and everything is is 10.9:
        if tag.endswith("_arm64"):
            return "macosx_11_0_arm64"

        return "macosx_10_9_x86_64"

    if tag.startswith("manylinux_"):
        # The v8 build process bundles its own Linux sysroots which work on Linuxes at
        # least this old (regardless of the platform we build on):
        if tag.endswith("_aarch64"):
            return "manylinux_2_27_aarch64"

        return "manylinux_2_27_x86_64"

    return tag


# From https://stackoverflow.com/questions/76450587/python-wheel-that-includes-shared-library-is-built-as-pure-python-platform-indep:
class PyMiniRacerBDistWheel(bdist_wheel):  # type: ignore[misc]
    def finalize_options(self) -> None:
        super().finalize_options()
        self.root_is_pure = False

    def get_tag(self) -> tuple[str, str, str]:
        return "py3", "none", _get_platform_tag()

    def run(self) -> None:
        mini_racer_src_dir = Path(__file__).parent / "src" / "py_mini_racer"
        if (
            not (mini_racer_src_dir / "mini_racer.dll").exists()
            and not (mini_racer_src_dir / "libmini_racer.so").exists()
            and not (mini_racer_src_dir / "libmini_racer.dylib").exists()
        ):
            # PyMiniRacer does not support a traditional from-source pip build, because
            # the v8 build is generally very slow, fragile, has many external
            # dependencies (exactly *which* depends on your system), and furthermore for
            # many systems the only viable build option is to cross-compile from
            # *another* system.
            # The intent of PyMiniRacer project is to build for all major platforms from
            # the GitHub home, and publish binary wheels to pip, so that you do not need
            # to build the wheel from source yourself.
            # If you want to build PyMiniRacer, you should use PyMiniRacer's build
            # system, starting with `just build-dll`, on a supported platform.
            # If you have an architecture that PyMiniRacer does not yet provide a wheel
            # for, consider contributing a pull request to add it at:
            # https://github.com/bpcreech/PyMiniRacer.
            msg = "Run `just build-dll` before building a PyMiniRacer wheel."
            raise RuntimeError(msg)

        super().run()


def _generate_readme() -> str:
    return "\n".join(
        [
            (Path(__file__).parent / "README.md")
            .read_text(encoding="utf-8")
            .replace(
                "(py_mini_racer.png)",
                "(https://github.com/bpcreech/PyMiniRacer/raw/main/py_mini_racer.png)",
            ),
            """
## Release history
""",
            "\n".join(
                (Path(__file__).parent / "HISTORY.md")
                .read_text(encoding="utf-8")
                .splitlines()[1:]
            ).replace("\n## ", "\n### "),
        ],
    )


setup(
    long_description=_generate_readme(),
    long_description_content_type="text/markdown",
    cmdclass={"bdist_wheel": PyMiniRacerBDistWheel},
)



from __future__ import annotations

import ctypes
import sys
from contextlib import ExitStack, contextmanager
from importlib import resources
from pathlib import Path
from sys import platform
from threading import Lock
from typing import TYPE_CHECKING, Protocol, cast

from py_mini_racer._exc import MiniRacerBaseException
from py_mini_racer._value_handle import RawValueHandle

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from py_mini_racer._value_handle import RawValueHandleType


def _get_lib_filename(name: str) -> str:
    """Return the path of the library called `name` on the current system."""
    if platform == "darwin":
        prefix, ext = "lib", ".dylib"
    elif platform == "win32":
        prefix, ext = "", ".dll"
    else:
        prefix, ext = "lib", ".so"

    return prefix + name + ext


MR_CALLBACK = ctypes.CFUNCTYPE(None, ctypes.c_uint64, RawValueHandle)


if TYPE_CHECKING:

    class MrCallback(Protocol):
        def __call__(
            self, callback_id: int, raw_val_handle: RawValueHandleType
        ) -> None: ...


class MrCallbackDecorator(Protocol):
    def __call__(self, cb: MrCallback) -> MrCallback: ...


mr_callback_func = cast("MrCallbackDecorator", MR_CALLBACK)


def _build_dll_handle(dll_path: Path) -> ctypes.CDLL:  # noqa: PLR0915
    handle = ctypes.CDLL(str(dll_path))

    handle.mr_init_v8.argtypes = [ctypes.c_char_p, ctypes.c_char_p]

    handle.mr_init_context.argtypes = [MR_CALLBACK]
    handle.mr_init_context.restype = ctypes.c_uint64

    handle.mr_eval.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
        ctypes.c_uint64,
    ]
    handle.mr_eval.restype = ctypes.c_uint64

    handle.mr_free_value.argtypes = [ctypes.c_uint64, RawValueHandle]

    handle.mr_alloc_int_val.argtypes = [ctypes.c_uint64, ctypes.c_int64, ctypes.c_uint8]
    handle.mr_alloc_int_val.restype = RawValueHandle

    handle.mr_alloc_double_val.argtypes = [
        ctypes.c_uint64,
        ctypes.c_double,
        ctypes.c_uint8,
    ]
    handle.mr_alloc_double_val.restype = RawValueHandle

    handle.mr_alloc_string_val.argtypes = [
        ctypes.c_uint64,
        ctypes.c_char_p,
        ctypes.c_uint64,
        ctypes.c_uint8,
    ]
    handle.mr_alloc_string_val.restype = RawValueHandle

    handle.mr_free_context.argtypes = [ctypes.c_uint64]

    handle.mr_context_count.argtypes = []
    handle.mr_context_count.restype = ctypes.c_size_t

    handle.mr_cancel_task.argtypes = [ctypes.c_uint64, ctypes.c_uint64]

    handle.mr_heap_stats.argtypes = [
        ctypes.c_uint64,
        ctypes.c_uint64,
    ]
    handle.mr_heap_stats.restype = ctypes.c_uint64

    handle.mr_low_memory_notification.argtypes = [ctypes.c_uint64]

    handle.mr_make_js_callback.argtypes = [
        ctypes.c_uint64,
        ctypes.c_uint64,
    ]
    handle.mr_make_js_callback.restype = RawValueHandle

    handle.mr_heap_snapshot.argtypes = [
        ctypes.c_uint64,
        ctypes.c_uint64,
    ]
    handle.mr_heap_snapshot.restype = ctypes.c_uint64

    handle.mr_get_identity_hash.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
    ]
    handle.mr_get_identity_hash.restype = RawValueHandle

    handle.mr_get_own_property_names.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
    ]
    handle.mr_get_own_property_names.restype = RawValueHandle

    handle.mr_get_object_item.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
        RawValueHandle,
    ]
    handle.mr_get_object_item.restype = RawValueHandle

    handle.mr_set_object_item.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
        RawValueHandle,
        RawValueHandle,
    ]
    handle.mr_set_object_item.restype = RawValueHandle

    handle.mr_del_object_item.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
        RawValueHandle,
    ]
    handle.mr_del_object_item.restype = RawValueHandle

    handle.mr_splice_array.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
        ctypes.c_int32,
        ctypes.c_int32,
        RawValueHandle,
    ]
    handle.mr_splice_array.restype = RawValueHandle

    handle.mr_call_function.argtypes = [
        ctypes.c_uint64,
        RawValueHandle,
        RawValueHandle,
        RawValueHandle,
        ctypes.c_uint64,
    ]
    handle.mr_call_function.restype = ctypes.c_uint64

    handle.mr_set_hard_memory_limit.argtypes = [ctypes.c_uint64, ctypes.c_size_t]

    handle.mr_set_soft_memory_limit.argtypes = [ctypes.c_uint64, ctypes.c_size_t]
    handle.mr_set_soft_memory_limit.restype = None

    handle.mr_hard_memory_limit_reached.argtypes = [ctypes.c_uint64]
    handle.mr_hard_memory_limit_reached.restype = ctypes.c_bool

    handle.mr_soft_memory_limit_reached.argtypes = [ctypes.c_uint64]
    handle.mr_soft_memory_limit_reached.restype = ctypes.c_bool

    handle.mr_v8_version.argtypes = []
    handle.mr_v8_version.restype = ctypes.c_char_p

    handle.mr_v8_is_using_sandbox.argtypes = []
    handle.mr_v8_is_using_sandbox.restype = ctypes.c_bool

    handle.mr_value_count.argtypes = [ctypes.c_uint64]
    handle.mr_value_count.restype = ctypes.c_size_t

    return handle


# V8 internationalization data:
_ICU_DATA_FILENAME = "icudtl.dat"

DEFAULT_V8_FLAGS = ("--single-threaded",)


class LibNotFoundError(MiniRacerBaseException):
    """MiniRacer-wrapped V8 build not found."""

    def __init__(self, path: Path) -> None:
        super().__init__(f"Native library or dependency not available at {path}")


class LibAlreadyInitializedError(MiniRacerBaseException):
    """MiniRacer-wrapped V8 build not found."""

    def __init__(self) -> None:
        super().__init__(
            "MiniRacer was already initialized before the call to init_mini_racer",
        )


def _open_resource_file(filename: str, exit_stack: ExitStack) -> Path:
    resource_path = resources.files("py_mini_racer") / filename

    context_manager = resources.as_file(resource_path)

    return exit_stack.enter_context(context_manager)


def _check_path(path: Path) -> None:
    if not path.exists():
        raise LibNotFoundError(path)


@contextmanager
def _open_dll(flags: Iterable[str]) -> Iterator[ctypes.CDLL]:
    dll_filename = _get_lib_filename("mini_racer")

    with ExitStack() as exit_stack:
        # Find the dll and its external dependency files:
        meipass = getattr(sys, "_MEIPASS", None)
        if meipass is not None:
            # We are running under PyInstaller.
            # See https://github.com/bpcreech/PyMiniRacer/issues/78
            meipass_path = Path(meipass)
            dll_path = meipass_path / "py_mini_racer" / dll_filename
            icu_data_path = meipass_path / "py_mini_racer" / _ICU_DATA_FILENAME
        else:
            dll_path = _open_resource_file(dll_filename, exit_stack)
            icu_data_path = _open_resource_file(_ICU_DATA_FILENAME, exit_stack)

        _check_path(dll_path)
        _check_path(icu_data_path)

        handle = _build_dll_handle(dll_path)

        handle.mr_init_v8(
            " ".join(flags).encode("utf-8"), str(icu_data_path).encode("utf-8")
        )

        yield handle


_init_lock = Lock()
_dll_handle_context_manager = None
_dll_handle = None


def init_mini_racer(
    *,
    flags: Iterable[str] = DEFAULT_V8_FLAGS,
    ignore_duplicate_init: bool = False,
) -> ctypes.CDLL:
    """Initialize py_mini_racer (and V8).

    This function can optionally be used to set V8 flags. This function can be called
    at most once, before any instances of MiniRacer are initialized. Instances of
    MiniRacer will automatically call this function to initialize MiniRacer and V8.
    """

    global _dll_handle_context_manager  # noqa: PLW0603
    global _dll_handle  # noqa: PLW0603

    with _init_lock:
        if _dll_handle is None:
            _dll_handle_context_manager = _open_dll(flags)
            _dll_handle = _dll_handle_context_manager.__enter__()
            # Note: we never call _dll_handle_context_manager.__exit__() because it's
            # designed as a singleton. But we could if we wanted to!
        elif not ignore_duplicate_init:
            raise LibAlreadyInitializedError

        return _dll_handle



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py_mini_racer._types import PythonJSConvertedTypes


class MiniRacerBaseException(Exception):  # noqa: N818
    """Base MiniRacer exception."""


class JSEvalException(MiniRacerBaseException):
    """JavaScript could not be executed."""


class JSTimeoutException(JSEvalException):
    """JavaScript execution timed out."""

    def __init__(self) -> None:
        super().__init__("JavaScript was terminated by timeout")


class JSPromiseError(MiniRacerBaseException):
    """JavaScript rejected a promise."""

    def __init__(self, reason: PythonJSConvertedTypes) -> None:
        super().__init__(f"JavaScript rejected promise with reason: {reason}\n")
        self.reason = reason


class JSArrayIndexError(IndexError, MiniRacerBaseException):
    """Invalid index into a JSArray."""

    def __init__(self) -> None:
        super().__init__("JSArray deletion out of range")
